// $Revision: 1.7 $
//
// EPSRevision History
//
//   J. Kercheval	 Mon, 09/16/1991 22:46:52	creation
//   J. Kercheval	 Sun, 09/22/1991 22:20:02	add automatic word file search and load
//   J. Kercheval	 Sun, 06/21/1992 00:29:35	modify for V6.0 highlighted regions
//
//   E. L. Olsen   4/21/93
//
//	   1) Set point to begining of newly entered words so that they
//	   are spell checked again, after they are entered.
//
//	   2) Made change so that the word currently being checked is
//	   hilighted, rather than displaying the percent complete.
//
//	   3) The ignore buffer loaded from file "ignore.lst", if it
//	   exists.
//
//	   4) Made ignore buffer visible so that it can be saved as a
//	   file.
//
//	   5) It used to be that all ignored words were added to the
//	   ignore list.  Now you must enter a space space to causes a
//	   word to be added to the ignore list.  Entering a carriage
//	   return still causes words to be ignored, but they are not
//	   added to the list.
//
//   J. Kercheval  Wed, 09/14/1994 12:54:05	Add modifications from Alan Thompson
//   J. Kercheval  Wed, 09/14/1994 13:26:08	Correct a problem with possessive checks
//
//   E. L. Olsen 12/19/94 Fixed bug in do_spell_region that caused word
//   not to be rechecked if user entered the same spelling.
//
//   E. L. Olsen 03/09/95 Fixed two bugs in spell_region: Test of
//   return value from lookpath wordfile was incorrect (it was
//   dereferenced) and call to gripe was incorrect.  Call to gripe was
//   replaced with call to say.
//
//   E. L. Olsen 06/06/95 Fixed a bug in do_spell_region that became
//   apparent in epsilon 7.0.  Replaced use of matchstart with mark.
//
//   A. Thompson 11/xx/96
//   
//	   1) Fixt bug in ignore look-up.  (Any sub-string in the ignore
//	   list would become a valid word.  i.e. add "noonish to the
//	   ignore list and "noo", "oon", "onis" and others such became
//	   valid words.)
//
//	   2) Added spell_word() and spell_AHD_word().  (I know.  I could
//	   have set-mark, forward-word and done a spell_region(); but I
//	   like it better this way.)  spell_AHD_word() gets the word at
//	   point to the echo area just as if it had been misspelled.  (If
//	   I'm sitting at something like "weather", spell_word() will
//	   tell me if it is in the dictionary but not if I really meant
//	   to use the word "whether".)
//
//	   3) I had put my ignore.lst in the same directory as words.lst
//	   but this thing wouldn't find it.  (Maybe, it was suppose to
//	   have an ignore list per directory so's to let the user have
//	   more flexibility.)  It now finds my ignore.lst.
//
//	   4) Changed it so as to give a running total in the echo area
//	   rather than highlighting the current word.  It takes too long
//	   the other way.
//
//	   5) Changed it back to having all correct words added to the
//	   ignore list.  (For words not in the dictionary or the ignore
//	   list, you get a choice.)  I like having a maintainable ignore
//	   list but the issue here is speed.  Like: a factor of two point
//	   five.  I'm a speed freak.  This way, one can have the best of
//	   both worlds.  (I'll admit that it was easier to maintain the
//	   ignore list when only pre-approved words got entered but the
//	   ignore list will become relativity stable in time and the
//	   factor of two point five will be with us forever.)
//
//	   6) Changed ignore.lst to -ignore.lst.  There are advantages to
//	   having the ignore list follow you around from session to
//	   session but there are disadvantages, too.  
//
//	   7) Changed the definition of a word.  It had been that a word
//	   could have numbers embedded in them and then a function was
//	   invokt to correct that.  That went.  Also, a word must be at
//	   least two characters long and apostrophes are valid in a word.
//	   (Now, you can add the word "fo'c's'le" to your dictionary.  It
//	   was such that I would reference something like "I.B.M." and
//	   this thing would stop at the "B" and the "M".  No more.)
//
//	   8) Added a new binary search routine.  See notes within
//	   B_SRCH_2.E for details.
//
//	   9) Changed to half of 132 column mode.  (The listing, I mean.
//	   I am most productive in 132 column mode, split down he middle.
//	   Especially when doing maintenance work.  Try it.)
//
//	   Note: The rules changed.  Both the dictionary and the ignore
//	   list must have a leading carriage return.  The dictionary has
//	   as its first entry: "a".  This will suffice.  If you have an
//	   ignore list, this routine won't see it because I've changed
//	   its name to "-ignore.lst".  Rename it and add a carriage
//	   return as the first character.
//
//   J. Kercheval  Thu, 06/26/1997 22:33:26	Incorporate changes by
//		E. L. Olsen and A. Thompson
// 
//
// 
//  This file implements a basic spelling checker which relies upon an
//  external list of correctly spelled, sorted words in an ascii
//  format and delimited by newlines. Case is ignored in this speller.
// 
//  The basic algorithm is brute force and simple marches through
//  every word in the input region (or buffer) and checks that word
//  against the wordlist.  If the word in not contained in the word
//  list then an ignore buffer which is being maintained is looked at
//  and if the word is not there then the user is presented with the
//  word and allowed to change it.  Entering a space will cause the
//  word to be placed in the ignore buffer for future occurances of
//  that word.  Entering just a carriage return will cause just the
//  current occurance of the word to be ignored.
// 
//  There are two commands implemented, spell_region and spell_buffer
//  (spell_buffer is just a call to spell_region with the values of
//  point and mark as 0 and size() respectively.  The search of the
//  word in the dictionary is done via a binary search.  This brute
//  force method is considerably slower than a trie or dictionary
//  method but is sufficient for a programmer's editor and is quite a
//  bit more flexible and easier to write.  A new buffer is created by
//  the name of "-words" which is not deleted upon completion to allow
//  for faster response on later spell checks.
// 
//  See the edit history for details and discussion of the ignore
//  buffer.
// 
// 
// 		John Kercheval
// 		4002 NE 204th St
// 		Lake Forest Park, WA  98155-1606
// 		INTERNET: kercheval@bigfoot.com, johnk@wrq.com
// 		Compu$erve: 72450,3702
// 
// 
//  Load binsrch2.b, then load spell.b.  Place a dictionary by the
//  name of WORDS.LST in your epsilon path (as signified by the
//  environment variable EPSPATH).  The same place as the file EDOC is
//  a good place to place the dictionary file.  Do you want to bind
//  these commands to a key (I did not want to so I left them out of
//  the function header)?  Now save the state file and you are set...
// 
// 								jbk
// 

#include <eel.h>

#ifndef BOOLEAN
#define BOOLEAN int
#define TRUE 1
#define FALSE 0
#endif

#define WORD_LIST_BUFFER	"-words"
#define WORD_FILE_NAME		"words.lst"

#define IGNORE_BUFFER		"-ignore"
#define IGNORE_FILE_NAME	"ignore.lst"

#define SPELL_WORD_STRING	"[a-zA-Z--][a-zA-Z--']+"

int size_of_dict;
int dict_sizer;
int spell_no_word = FALSE;


////////////////////////////////////////////////////////////////////////////
//
// do_spell_region() sets up the needed buffers initializes the point
// in the input buffer and repeatedly calls spell_check to actually
// perform the binary search for the element
//

do_spell_region(inbuf, wordbuf, ignorebuf)
char *inbuf;
char *wordbuf;
char *ignorebuf;
{
	char elem_3[256];
	char elem_2[256];
	char elem[256];
	char res[256];
	char message[256];

	spot end = alloc_spot();

	jmp_buf *old_level = top_level;
	jmp_buf this_level;
	int ret;
								   
	int num_checked = 0;

	//
	// store the old point, mark and buffer, store and turn of
	// highlight
	//
	save_spot point, mark;
	save_var bufname;
	save_var _highlight_control = 0;

	//
	// clean up if abort key is pressed or error occurs while
	// processing.
	//
	top_level = &this_level;
	if (ret = setjmp(top_level)) {
		free_spot(end);
		top_level = old_level;
		longjmp(top_level, ret);
	}

	//
	// move to buffer
	//
	bufname = inbuf;

	//
	// set point and end location
	//
	if (point > mark) {
		*end = point;
		point = mark;
	} else {
		*end = mark;
	}
	note("Working...  Checked %d items. ", num_checked);
	highlight_on ();

	//
	// run through the region
	//
	while (point < *end) {
		//
		// get next word to check
		//
		if (!re_search(1, SPELL_WORD_STRING)) {
			break;
		}
		num_checked++;

		grab(matchstart, point, elem);
		strcpy (elem_2, elem);
		strcat (elem_2, "\n\0");
		strcpy (elem_3, "\n");
		strcat (elem_3, elem_2);

		//
		// begin highlight the region (and thus the word)
		// 
		mark = matchstart;
		note("Working...  Checked %d items. ",  num_checked);

		// 
		// see if this is an AHD look-up
		//
		if (spell_no_word) {
			//
			// Just get the word under the current point
			//
			spell_no_word = FALSE;
			refresh();
			get_strdef(res, "Enter change or <CR> to continue. ", elem);
			if (strfcmp(elem, res)) {
				delete(mark, point);
				stuff(res);
				refresh();
			}
			continue;
		}

		//
		// check ignore list
		//
		bufname = ignorebuf;
		point = 0;
		if (search(1, elem_3)) {
			bufname = inbuf;
			continue;
		}

		//
		// check dictionary
		//
		bufname = wordbuf;
		if (do_binary_search(elem, size_of_dict, dict_sizer)) {
			bufname = ignorebuf;
			point = 1;
			stuff(elem);
			stuff("\n");
			bufname = inbuf;
			continue;
		}

		//
		// If this is a possessives, lose the
		// "apostrophe s". If it is a close quote,
		// lose that; but not both.
		//
		bufname = inbuf;
		if ((character(point-1) == 's') ||
			(character(point-1) == 'S')) {
			if (character(point-2) == '\x27') {
				point -=2;
			}
		} else {
			if (character(point-1) == '\x27') {
				point--;
			}
		}

		//
		// regrab the word after apostrophe check
		//
		grab(mark, point, elem);
		strcpy (elem_2, elem);
		strcat (elem_2, "\n\0");
		strcpy (elem_3, "\n");
		strcat (elem_3, elem_2);
		
		//
		// check ignore list, again
		//
		bufname = ignorebuf;
		point = 0;
		if (search(1, elem_3)) {
			bufname = inbuf;
			continue;
		}
		
		//
		// check dictionary, again
		//
		bufname = wordbuf;
		if (do_binary_search(elem, size_of_dict, dict_sizer)) {
			bufname = ignorebuf;
			point = 1;
			stuff(elem);
			stuff("\n");
			bufname = inbuf;
			continue;
		}
		
		//
		// get user input
		//
		refresh();
		bufname = inbuf;
		if (insert_default_response) {
			strcpy (message, "<CR>=add, <Delete><CR>=ignore; or change: ");
		} else {
			strcpy (message, "<CR>=add, ^S<CR>=ignore; or change. [");
			strcat (message, elem);
			strcat (message, "]: \0");
		}
		get_strnone(res, message, elem);
		if (strlen(res) &&
			strfcmp(elem, res)) {
			//
			// user changed the elem
			//
			delete(mark, point);
			stuff(res);

			//
			// set point to beginning of user's new word
			//
			point = mark;
			highlight_on ();
			continue;
		}
		//
		// At this point the response is of zero length or identical
		// to the string.
		//
		if (strlen(res) &&
			!insert_default_response) {
			//
			// The user said to just forget about this one by doing
			// ^S<CR> 
			//
			continue;
		} else if (!strlen(res) &&
				   insert_default_response) {
			//
			// The user said to just forget about this one by doing
			// <Delete><CR> 
			//
			continue;
		} else {
			//
			// ignore all instances of this word (user just hit <CR>)
			//
			bufname = ignorebuf;
			point = 1;
			stuff(elem);
			stuff("\n");
			bufname = inbuf;
		}
		note("Working...  Checked %d items. ",  num_checked);

	} // end while
	
	//
	// clean up
	//
	note("Done.  Checked %d items.", num_checked);
	free_spot(end);
	top_level = old_level;
}


////////////////////////////////////////////////////////////////////////////
//
// spell_region() and spell_buffer() are the top level commands.
// They call do_spell_region() to do all of the work.
//

command spell_region()
{
	char *oldbuf = bufname;
	char *words;
	
	int l_size_of_dict;
	int dict_multiplier;
	int file_error;

	note("Loading dictionary...");

	if (!exist(WORD_LIST_BUFFER)) {
		//
		// obtain the word file
		//
		words = lookpath(WORD_FILE_NAME);
		if (words) {
			create(WORD_LIST_BUFFER);
			bufname = WORD_LIST_BUFFER;
			file_error = file_read(words, FILETYPE_AUTO);
			if (file_error)	{
				say("Error reading word file.");
				bufname = oldbuf;
				return;
			}
			size_of_dict = size();
			l_size_of_dict = 1;
			dict_multiplier = 2;
			while (l_size_of_dict < size_of_dict) {
				l_size_of_dict += dict_multiplier;
				dict_multiplier = (dict_multiplier << 1);
			}
			
			//
			// l_size_of_dict is now larger than the size
			// of the dictionary and of the form: (2**n-1).
			// Make dict_sizer of the form: (2**n).  See
			// notes in b_srch_2.e for details.
			//
			dict_sizer = l_size_of_dict + 1;
			bufname = oldbuf;
		} else {
			say("Word file '%s' not found in EPSPATH", WORD_FILE_NAME);
			return;
		}
	}

	note("Loading ignore.lst...");

	if (!exist(IGNORE_BUFFER)) {
		//
		// create or obtain ignore file
		//
		create(IGNORE_BUFFER);
		words = lookpath(IGNORE_FILE_NAME);
		if (words) {
			bufname = IGNORE_BUFFER;
			file_error = file_read(words, FILETYPE_AUTO);
			if (file_error) {
				say("Error reading ignore file.");
				zap(IGNORE_BUFFER);
				stuff("\n");
				bufname = oldbuf;
			}
		} else {
			bufname = IGNORE_BUFFER;
			stuff("\n");
		}
		bufname = oldbuf;
	}
	//
	// run the spell checker
	//
	do_spell_region(bufname, WORD_LIST_BUFFER, IGNORE_BUFFER);
}

command spell_buffer()
{
    save_spot point, mark;

	//
	// set the region and run the spell checker
	//
    mark = 0;
    point = size();
    spell_region();
}

command spell_word()
{
    save_spot point, mark;
	//
	// set point and mark
	//
    point -=2;
    re_search(1, SPELL_WORD_STRING);
    mark = point;
    re_search(-1, SPELL_WORD_STRING);
    point = matchstart;
    mark = matchend;
    spell_region();
}

command spell_AHD_word()
{
	//
	// set a switch and invoke spell_word()
	//
    spell_no_word = TRUE;
    spell_word();
}

